
/**************************************************************
 *
 * Unix i C
 * (C) Pawel Gorecki 2005
 * gorecki@mimuw.edu.pl
 * 
 */

#ifndef _MMEM_H_
#define _MMEM_H_

#include <string.h>
#include <malloc.h>

/***********************************************************
 * 
 * macro: _MEMCHK_
 *   minf() - malloc info
 *   mchk() - memory consistency check
 *
 * macro: _MEMCONTROL_
 *   mmalloc(size) - malloc
 *   mstrdup(s)    - strdup
 *   mfree(p)      - free
 *   ppMem()       - print mem info
 *
 ***********************************************************/ 

#ifdef __GNUG__
#define EXTERNC extern "C"
#else
#define EXTERNC
#endif

#define tmalloc(cname) ((cname*)mmalloc(sizeof(cname)))

/*********************************************************/ 

#ifdef _MEMCHK_

#include <mcheck.h> 

EXTERNC void _checkinf(char *proc, int l, char *f, int flag);

#define minf() _ppinfo(__LINE__,__FILE__)
#define mchk() _myMemChecking(__LINE__,__FILE__)

EXTERNC void _ppinfo(int l,char *f);
EXTERNC void _myMemChecking(int l,char *f);

#else

#define minf()
#define mchk()

#endif

/*********************************************************/

#ifdef _MEMCONTROL_

typedef struct memEntry
{
    void *p;
    int size;
    char *file;
    int linenum;
    int status;
    struct memEntry *next;
    char *info;
} memEntry;

#define mmalloc(s) _mmalloc(__LINE__,__FILE__,s,#s)
#define mstrdup(s) _mstrdup(__LINE__,__FILE__,s,#s)
#define mfree(p) _mfree(__LINE__,__FILE__,p)
#define ppMem() _ppMem(__LINE__,__FILE__)

EXTERNC void *_mmalloc(int l, char *f, int size,char *info);
EXTERNC char *_mstrdup(int l, char *f, char *s, char *info);
EXTERNC void _mfree(int l, char *f, void *p);
EXTERNC void _ppMem(int l,char *f);

#else

#define mmalloc(s) malloc(s)
#define mstrdup(s) (char*)strdup(s)
#define mfree(p) free(p)
#define ppMem() 

#define _mmalloc(l,f,s,info) malloc(s)
#define _mstrdup(l,f,s,info) strdup(s)
#define _mfree(l, f, p) free(p)
#define _ppMem(l,f)

#endif

#endif

