
/**************************************************************
 *
 * Unix i C
 * (C) Pawel Gorecki 2005
 * gorecki@mimuw.edu.pl
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <mmem.h>

#define STRDUP 1
#define MALLOC 2
#define FREE 4
#define SUNUSED 4

int ods=0;

struct mallinfo prevmi;
int initialized=0;

void _checkinf(char *proc, int l, char *f, int flag)
{
  struct mallinfo m;
  if (!initialized)
    {
      prevmi = mallinfo();
      initialized=1;
      return;
    }
  m=mallinfo();
  
#define MAXINCMEM 20000000

#define WARN(str) { fprintf(stderr,"%s:%d: Warning %s %s %s\n",f,l,str,(!flag)?"before":"after",proc);  minf(); \
            fprintf(stderr,"%d %d, %d %d, %d %d\n",	\
            m.arena,    prevmi.arena, 	  \
            m.uordblks, prevmi.uordblks,	  \
            m.fordblks, prevmi.fordblks);}
  if (m.arena != (m.uordblks + m.fordblks)) WARN("illegal sizes");
  if (((labs(prevmi.arena - m.arena)) > MAXINCMEM) 
      || ((labs(prevmi.uordblks - m.uordblks)) > MAXINCMEM)
      || ((labs(prevmi.fordblks - m.fordblks)) > MAXINCMEM)
      )
    WARN("strange de/increments");
  prevmi=m;
}

memEntry *allocList = NULL;
int allocListlen = 0;

void *_mmalloc(int l, char *f, int size, char *info)
{
  
  memEntry *c = (memEntry*)malloc(sizeof(memEntry));
  _checkinf("malloc",l,f,0);
  if (!c) 
    {
      fprintf(stderr,"Can't allocate memEntry (malloc). File %s line %d <%s>\n",f,l,info);
      fflush(stderr);
      exit(-1);
    }
  c->p = malloc(size);
  if (!c) 
    {
      fprintf(stderr,"Can't allocate mem size %d in file %s line %d <%s>\n",size,f,l,info);
      fflush(stderr);
      exit(-1);
    }

#ifdef MDEB
  fprintf(stderr,"Inserting mmalloc  %d %s %d %s\n",l,f,size,info);
#endif
  
  c->size = size;
  c->linenum = l;
  c->file=f;
  c->status = MALLOC;
  c->info = info;
  c->next = allocList;
  allocListlen++;
  allocList = c;
  _checkinf("malloc",l,f,1);  
  return c->p;
}

char* _mstrdup(int l, char *f, char *s,char *info)
{  
  memEntry *c;
  _checkinf("strdup", l,f,0);
  c = (memEntry*)malloc(sizeof(memEntry));
  if (!c) 
    {
      fprintf(stderr,"Can't allocate memEntry (strdup). File %s line %d <%s>\n",f,l,info);
      fflush(stderr);
      exit(-1);
    }
  c->p = strdup(s);
  if (!c) 
    {
      fprintf(stderr,"Can't allocate str. Size %d in file %s line %d <%s>\n",strlen(s),f,l,info);
      fflush(stderr);
      exit(-1);
    }

#ifdef MDEB
  fprintf(stderr,"Inserting strdup  %d %s %s %s\n",l,f,s,info);
#endif
  
  c->size = strlen(s)+1;
  c->linenum = l;
  c->file=f;
  c->status = STRDUP;
  c->next = allocList;
  c->info=info;
  allocListlen++;
  allocList = c;  
  _checkinf("strdup",l,f,1);
  return (char*)c->p;
}

void _mfree(int l, char *f, void *p)
{
  memEntry *ls = allocList;
  memEntry *prev = NULL;
  _checkinf("free",l,f,0);

#ifdef MDEB
  fprintf(stderr,"Free  %d %s\n",l,f);
#endif

  while (ls)
    {
      if ((ls->p==p) && 
	  ((ls->status==MALLOC) || (ls->status==STRDUP)))
	{	
	  if (prev==NULL)
	    allocList = ls->next;
	  else
	    prev->next=ls->next;
	  free(ls);	      
	  free(p);
	  return;
	}
      prev=ls;
      ls=ls->next;
    }
  fprintf(stderr,"Error (lst)! Illegal free in %s line %d ",f,l);
  fprintf(stderr,"free(.) - not executed.\n");  
  _checkinf("free",l,f,1);
}

void _ppMem(int ln, char *f)
{
  int total=0;
  int pos=0;
  memEntry *l = allocList;
  int c=0;
  fprintf(stderr,"%s:%d: Memory allocation info\n",f,ln);
  while (l)
    {
      if (l->status & FREE)
	{ 
	  //	  fprintf(stderr," (FREE) ");	  
	}
      else
	{
	  pos++;
	  fprintf(stderr,"%d. %s:%d",c,l->file,l->linenum);
	  if (l->info)
	    fprintf(stderr," INFO=<%s>",l->info);
	  if (!(l->status & FREE)) 
	    {
	      if (l->status & STRDUP)
		{
		  fprintf(stderr," STR=<%s>",(char*)l->p);
		  total+=strlen((char*)l->p);
		}
	      if (l->status & MALLOC)
		{
		  fprintf(stderr," PTR size=<%d>",l->size);
		  total+=l->size;
		}
	      fprintf(stderr,"\n");
	    }
	}
      c++;
      l=l->next;
    }
  fprintf(stderr,"Total: %dB, %d pos.\n",total,pos);
}

